#include <stdio.h>

#include "cuda_utils.h"

const int blockSizeX=128;
const int blockSizeY=4;

texture<float,2> texRef;

void FdtdStepReference(int sizeX,int sizeY,float* dataOut,float* dataIn) {
  for(int i=1;i<sizeY-1;i++) {
    for(int j=1;j<sizeX-1;j++) {
      int index=i*sizeX+j;
      dataOut[index] = 0.5f * dataIn[index] +
        0.125f * (   dataIn[index-1]     + dataIn[index+1]
                   + dataIn[index-sizeX] + dataIn[index+sizeX]);
    }
  }
}

void __global__ FdtdStep(int sizeX,int sizeY,float* dataOut,float* dataIn)
{
  int myX=blockIdx.x*blockSizeX+threadIdx.x;
  int myY=blockIdx.y*blockSizeY+threadIdx.y;
  if (myX==0 || myX>=sizeX-1 || myY==0 || myY>=sizeY-1)
    return;
  int myIndex=myY*sizeX+myX;
  
  // TODO: replace all references to data in dataIn with tex2D calls
  // WARNING: do not change the order of the operands in the parentheses
  // '(...+...+...+)'. 
  // Algorithm is numerically unstable -> result check might fail(!)
  dataOut[myIndex] = 0.5f*dataIn[myIndex] +
    0.125f * (   dataIn[myIndex-1]     + dataIn[myIndex+1]
               + dataIn[myIndex-sizeX] + dataIn[myIndex+sizeX] );
} 

int main()
{
  // define field size //
  const int sizeX = 512;
  const int sizeY = 512;
  const int loops = 600;

  dim3 threads(blockSizeX,blockSizeY);
  dim3 grid((sizeX+blockSizeX-2)/blockSizeX,(sizeY+blockSizeY-2)/blockSizeY);

  float* dataA;
  float* dataB;
  float* dataDeviceA;
  float* dataDeviceB;

  int dataSize=sizeof(float)*sizeX*sizeY;

  // allocate memory on cpu //
  cu_verify(cudaMallocHost((void**)&dataA,dataSize));
  cu_verify(cudaMallocHost((void**)&dataB,dataSize));
  
  // initialize fields //
  for(int i=0;i<sizeY;i++) {
    for(int j=0;j<sizeX;j++) {
      dataA[i*sizeX+j]=1.f;
      dataB[i*sizeX+j]=1.f;
    }
  }
  dataA[(sizeY/2)*sizeX+(sizeX/2)]=2.f;

  // allocate memory on gpu //
  cu_verify(cudaMalloc((void**)&dataDeviceA,dataSize));
  cu_verify(cudaMalloc((void**)&dataDeviceB,dataSize));

  // copy input data from cpu to gpu //
  cu_verify(cudaMemcpy(dataDeviceA,dataA,dataSize,cudaMemcpyHostToDevice));
  cu_verify(cudaMemcpy(dataDeviceB,dataB,dataSize,cudaMemcpyHostToDevice));

  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float>();

  cu_verify(cudaThreadSynchronize());

  Timer timer;
  initTimer(&timer);
  // computation main loop //
  for(int32_t i=0;i<loops;i++) {

    // bind texture to array A //
    // TODO: add call to cudaBindTexture2D

    // call kernel //
    cu_verify_void((FdtdStep<<< grid,threads >>>(sizeX,sizeY,
                                                 dataDeviceB,dataDeviceA)));

    // bind texture to array B //
    // TODO: add call to cudaBindTexture2D

    // call kernel //
    cu_verify_void((FdtdStep<<< grid,threads >>>(sizeX,sizeY,
                                                 dataDeviceA,dataDeviceB)));
  }
  cudaThreadSynchronize();
  double duration=getTimer(&timer);
  printf("Time=%e s, speed=%e flops\n",duration,
         12.*loops*(sizeX-2)*(sizeY-2)/duration);

  // copy results from gpu to cpu //
  cu_verify(cudaMemcpy(dataA,dataDeviceA,dataSize,cudaMemcpyDeviceToHost));
  cu_verify(cudaMemcpy(dataB,dataDeviceB,dataSize,cudaMemcpyDeviceToHost));

  // verify results
  float* dataReferenceA = (float*)malloc(dataSize);
  float* dataReferenceB = (float*)malloc(dataSize);
  for(int i=0;i<sizeY;i++) {
    for(int j=0;j<sizeX;j++) {
      dataReferenceA[i*sizeX+j]=1.f;
      dataReferenceB[i*sizeX+j]=1.f;
    }
  }
  dataReferenceA[(sizeY/2)*sizeX+(sizeX/2)]=2.f;
  for(int32_t i=0;i<loops;i++) {
    FdtdStepReference(sizeX,sizeY,dataReferenceB,dataReferenceA);
    FdtdStepReference(sizeX,sizeY,dataReferenceA,dataReferenceB);
  }
  bool success=true;
  for(int i=0;i<sizeY;i++) {
    for(int j=0;j<sizeX;j++) {
      /* if (fabsf(dataReferenceA[i*sizeX+j]-dataA[i*sizeX+j]) */
      /*     >1e-11*(fabsf(dataReferenceA[i*sizeX+j])+fabsf(dataA[i*sizeX+j])) */
      /*     && fabsf(dataA[i*sizeX+j])>1e-50) { */
      if (dataReferenceA[i*sizeX+j]!=dataA[i*sizeX+j] && fabs(dataReferenceA[i*sizeX+j])>1e-35) {
        printf("computation result is wrong: position=(%i,%i) reference=%.12e"
               " cuda result=%.12e\n",
               i,j,dataReferenceA[i*sizeX+j],dataA[i*sizeX+j]);
        printf("position=(%i,%i) reference=%.12e"
               " cuda result=%.12e\n",
               i,j,dataReferenceB[(i)*sizeX+j],dataB[(i)*sizeX+j]);
        printf("position=(%i,%i) reference=%.12e"
               " cuda result=%.12e\n",
               i-1,j,dataReferenceB[(i-1)*sizeX+j],dataB[(i-1)*sizeX+j]);
        printf("position=(%i,%i) reference=%.12e"
               " cuda result=%.12e\n",
               i,j-1,dataReferenceB[(i)*sizeX+j-1],dataB[(i)*sizeX+j-1]);
        printf("position=(%i,%i) reference=%.12e"
               " cuda result=%.12e\n",
               i,j+1,dataReferenceB[(i)*sizeX+j+1],dataB[(i)*sizeX+j+1]);
        printf("position=(%i,%i) reference=%.12e"
               " cuda result=%.12e\n",
               i+1,j,dataReferenceB[(i+1)*sizeX+j],dataB[(i+1)*sizeX+j]);
        success=false;
        break; 
      }
    }
    if (!success)
      break;
  }
  if (success) {
    printf("computation result is correct.\n");
  }
  free(dataReferenceA);
  free(dataReferenceB);

  // free cpu memory //
  cu_verify(cudaFreeHost(dataA));
  cu_verify(cudaFreeHost(dataB));

  // free gpu memory //
  cu_verify(cudaFree(dataDeviceA));
  cu_verify(cudaFree(dataDeviceB));

  // exit //
  return 0;
}
