#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "cuda_utils.h"

// CUDA Kernel //

void __global__ VectorSum(float* a,float* b,float* c,int size) {
  for(int i=threadIdx.x+blockIdx.x*blockDim.x; i<size; 
      i+=blockDim.x*gridDim.x) {
    c[i]=a[i]+b[i];
  }
}


// CPU code //

int main() {

  srand48(time(NULL));

  // define vector size //
  const int vector_size = 10000000;

  // allocate memory on cpu //
  float* vector_a_cpu;
  float* vector_b_cpu;
  float* vector_c_cpu;
  cu_verify(cudaMallocHost((void**)&vector_a_cpu,vector_size*sizeof(float)));
  cu_verify(cudaMallocHost((void**)&vector_b_cpu,vector_size*sizeof(float)));
  cu_verify(cudaMallocHost((void**)&vector_c_cpu,vector_size*sizeof(float)));

  // initialize host vectors //
  for(int i=0;i<vector_size;i++) {
    vector_a_cpu[i]=drand48();
    vector_b_cpu[i]=drand48();
    vector_c_cpu[i]=0;
  }

  // allocate memory on gpu //
  float* vector_a_gpu;
  float* vector_b_gpu;
  float* vector_c_gpu;
  cu_verify(cudaMalloc((void**)&vector_a_gpu,vector_size*sizeof(float)));
  cu_verify(cudaMalloc((void**)&vector_b_gpu,vector_size*sizeof(float)));
  cu_verify(cudaMalloc((void**)&vector_c_gpu,vector_size*sizeof(float)));

  // copy input vectors from cpu to gpu //
  cu_verify(cudaMemcpy(vector_a_gpu,vector_a_cpu,vector_size*sizeof(float),
                       cudaMemcpyHostToDevice));
  cu_verify(cudaMemcpy(vector_b_gpu,vector_b_cpu,vector_size*sizeof(float),
                       cudaMemcpyHostToDevice));
  // initialize gpu output vector to 0 //
  cu_verify(cudaMemcpy(vector_c_gpu,vector_c_cpu,vector_size*sizeof(float),
                       cudaMemcpyHostToDevice));

  Timer timer;
  initTimer(&timer);

  // cuda kernel call //
  cu_verify_void((VectorSum<<<128,64>>>(vector_a_gpu,vector_b_gpu,
                                        vector_c_gpu,vector_size)));

  cu_verify(cudaThreadSynchronize());
  double duration = getTimer(&timer);

  // copy output vector from gpu to cpu //
  cu_verify(cudaMemcpy(vector_c_cpu,vector_c_gpu,vector_size*sizeof(float),
                       cudaMemcpyDeviceToHost));

  // verify results //
  bool success=true;
  for(int i=0;i<vector_size;i++) {
    if (fabsf(vector_c_cpu[i]-(vector_a_cpu[i]+vector_b_cpu[i]))>1e-5) {
      printf("computation result is wrong: index=%i a[i]+b[i]=%g"
             " cuda result=%g\n",
             i,vector_a_cpu[i]+vector_b_cpu[i],vector_c_cpu[i]);
      success=false;
      break;
    }
  }
  if (success) {
    printf("computation result is correct.\n");
  }

  printf("duration=%le seconds, performance = %le FLOPS\n"
         "memory bandwidth = %le bytes/s\n",
         duration,vector_size/duration,vector_size*12./duration);

  // free gpu memory //
  cu_verify(cudaFree(vector_a_gpu));
  cu_verify(cudaFree(vector_b_gpu));
  cu_verify(cudaFree(vector_c_gpu));

  // free cpu memory //
  cu_verify(cudaFreeHost(vector_a_cpu));
  cu_verify(cudaFreeHost(vector_b_cpu));
  cu_verify(cudaFreeHost(vector_c_cpu));

  // exit //
  return 0;
}
