#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "cuda_utils.h"

// block size 'T' (width and height of the shared memory block) //
const int blockSize=16;

__global__ void MatrixProd(float* a,float* b,float* c, int m,int n,int p) {

  __shared__ float shared_a[blockSize][blockSize];
  __shared__ float shared_b[blockSize][blockSize];

  int index_i=blockIdx.y*blockSize+threadIdx.y;
  int index_j=blockIdx.x*blockSize+threadIdx.x;

  float sum=0;
  // iterate over blocks
  for(int index_k_coarse=0;index_k_coarse<n;index_k_coarse+=blockSize) {
    // read block of a into shared memory
    int column_a=threadIdx.x+index_k_coarse;
    if (column_a<n && index_i<m) {
      shared_a[threadIdx.y][threadIdx.x]=a[index_i*n+column_a];
    }
    // read block of b into shared memory
    int row_b=threadIdx.y+index_k_coarse;
    if (index_j<p && row_b<n) {
      shared_b[threadIdx.y][threadIdx.x]=b[row_b*p+index_j];
    }
    __syncthreads();
    // compute sum for one block
    if (index_j<p && index_i<m) {
      int end=min(blockSize,n-index_k_coarse);
      for(int index_k_fine=0;index_k_fine<end;index_k_fine++) {
        sum+=shared_a[threadIdx.y][index_k_fine]
             * shared_b[index_k_fine][threadIdx.x];
      }
    }
    __syncthreads();
  }
  // assign result to matrix c
  if (index_j<p && index_i<m) {
    c[index_i*p+index_j]=sum;
  }
}

// CPU code fore reference computation //

void MatrixProdReference(float* a,float* b,float* c,int m,int n,int p) {
  for(int index_i=0;index_i<m;index_i++) {
    size_t a_line_index=index_i*n;
    for(int index_j=0;index_j<p;index_j++) {
      float sum=0;
      for(int index_k=0;index_k<n;index_k++) {
        sum+=a[a_line_index+index_k]*b[index_k*p+index_j];
      }
      c[index_i*p+index_j]=sum;
    }
  }
}

int main() {

  srand48(time(NULL));

  // define sizes //
  const int size_m = 450;
  const int size_n = 350;
  const int size_p = 400;

  const size_t matrix_a_size = size_m*size_n*sizeof(float);
  const size_t matrix_b_size = size_n*size_p*sizeof(float);
  const size_t matrix_c_size = size_m*size_p*sizeof(float);

  // allocate memory on cpu //
  float* matrix_a_cpu;
  float* matrix_b_cpu;
  float* matrix_c_cpu;
  cu_verify(cudaMallocHost((void**)&matrix_a_cpu,matrix_a_size));
  cu_verify(cudaMallocHost((void**)&matrix_b_cpu,matrix_b_size));
  cu_verify(cudaMallocHost((void**)&matrix_c_cpu,matrix_c_size));
  
  // initialize matrices //
  for(int i=0;i<size_m*size_n;i++) {
    matrix_a_cpu[i]=drand48();
  }
  for(int i=0;i<size_n*size_p;i++) {
    matrix_b_cpu[i]=drand48();
  }
  for(int i=0;i<size_m*size_p;i++) {
    matrix_c_cpu[i]=-1;
  }

  // allocate memory on gpu //
  float* matrix_a_gpu;
  float* matrix_b_gpu;
  float* matrix_c_gpu;
  cu_verify(cudaMalloc((void**)&matrix_a_gpu,matrix_a_size));
  cu_verify(cudaMalloc((void**)&matrix_b_gpu,matrix_b_size));
  cu_verify(cudaMalloc((void**)&matrix_c_gpu,matrix_c_size));

  // initialize output matrix C on gpu //
  cu_verify(cudaMemcpy(matrix_c_gpu,matrix_c_cpu,matrix_c_size,
                       cudaMemcpyHostToDevice));

  // copy input matrices A, B from cpu to gpu //

  Timer timer_transfer_in;
  initTimer(&timer_transfer_in);

  cu_verify(cudaMemcpy(matrix_a_gpu,matrix_a_cpu,matrix_a_size,
                       cudaMemcpyHostToDevice));
  cu_verify(cudaMemcpy(matrix_b_gpu,matrix_b_cpu,matrix_b_size,
                       cudaMemcpyHostToDevice));

  dim3 threads(blockSize,blockSize);
  dim3 grid((size_p+blockSize-1)/blockSize,(size_m+blockSize-1)/blockSize);

  Timer timer_kernel;
  initTimer(&timer_kernel);

  // kernel call //
  cu_verify_void((MatrixProd<<<grid,threads>>>(matrix_a_gpu,matrix_b_gpu,
                                               matrix_c_gpu,
                                               size_m,size_n,size_p)));
  cu_verify(cudaThreadSynchronize());

  Timer timer_transfer_out;
  initTimer(&timer_transfer_out);

  // copy output vector from gpu to cpu //
  cu_verify(cudaMemcpy(matrix_c_cpu,matrix_c_gpu,matrix_c_size,
                       cudaMemcpyDeviceToHost));

  Timer timer_end;
  initTimer(&timer_end);

  // calculate reference values
  float* matrix_c_reference = (float*)malloc(matrix_c_size);

  Timer timer_cpu_begin;
  initTimer(&timer_cpu_begin);

  MatrixProdReference(matrix_a_cpu,matrix_b_cpu,matrix_c_reference,
                      size_m,size_n,size_p);

  Timer timer_cpu_end;
  initTimer(&timer_cpu_end);

  // verify results
  bool success=true;
  for(int i=0;i<size_m;i++) {
    for(int j=0;j<size_p;j++) {
      if (fabs(matrix_c_cpu[i*size_p+j]-matrix_c_reference[i*size_p+j])
          > 1e-5 * 
          fabs(matrix_c_cpu[i*size_p+j]+matrix_c_reference[i*size_p+j])) {
        fprintf(stderr,"error: position=%i,%i reference=%e cuda result=%e\n",
                i,j,matrix_c_reference[i*size_p+j],matrix_c_cpu[i*size_p+j]);
        success=false;
        break;
      }
    }
    if (!success)
      break;
  }

  free(matrix_c_reference);

  if (success) {
    printf("computation result is correct.\n");
  }

  double duration = getTimerDifference(&timer_transfer_in,&timer_kernel);
  fprintf(stderr,"duration (data transfer in) = %e s, bandwith = %e bytes/s\n",
          duration,(matrix_a_size+matrix_b_size)/duration);
  duration = getTimerDifference(&timer_transfer_out,&timer_end);
  fprintf(stderr,"duration (data transfer out) = %e s, bandwith = %e bytes/s\n",
          duration,matrix_c_size/duration);
  duration = getTimerDifference(&timer_kernel,&timer_transfer_out);
  fprintf(stderr,"duration (computation) = %e s, performance = %e FLOPS\n",
          duration,2.*size_m*size_n*size_p/duration);
  duration = getTimerDifference(&timer_transfer_in,&timer_end);
  fprintf(stderr,"duration (overall) = %e s, performance (including transfer time) = %e FLOPS\n",
          duration,2.*size_m*size_n*size_p/duration);
  duration = getTimerDifference(&timer_cpu_begin,&timer_cpu_end);
  fprintf(stderr,"duration (computation on CPU) = %e s, performance = %e FLOPS\n",
          duration,2.*size_m*size_n*size_p/duration);

  // free cgu memory //
  cu_verify(cudaFree(matrix_a_gpu));
  cu_verify(cudaFree(matrix_b_gpu));
  cu_verify(cudaFree(matrix_c_gpu));

  // free cpu memory //
  cu_verify(cudaFreeHost(matrix_a_cpu));
  cu_verify(cudaFreeHost(matrix_b_cpu));
  cu_verify(cudaFreeHost(matrix_c_cpu));

  // exit //
  return 0;
}
