module device_code

integer, parameter :: iterations=20

contains

! CUDA Kernel !

attributes(global) subroutine compute(input,output,size)
  implicit none
  real, dimension(0:size-1) :: input,output
  integer, value :: size
  integer :: i,j
  real :: value
  do i=threadIdx%x-1+(blockIdx%x-1)*blockDim%x,size-1,blockDim%x*gridDim%x
    value=input(i)
    do j=1,iterations
      value=4.*value*(1.-value)
    end do
    output(i)=value
  end do
end subroutine compute

end module device_code

module cpu_code

use device_code

contains 

! iterate function for computation of reference !

real function iterate(value) 
  implicit none
  real :: value
  integer :: j
  do j=1,iterations
    value=4.*value*(1.-value)
  end do
  iterate=value
end function iterate

end module cpu_code

program exercise11

use cudafor
use cpu_code
use device_code

implicit none

! define vector size !
integer, parameter :: vector_size = 10000000
integer, parameter :: stream_count = 8
real, dimension(:),allocatable,pinned :: vector_input_cpu
real, dimension(:),allocatable,pinned :: vector_output_cpu
real, dimension(0:vector_size-1), device :: vector_input_gpu
real, dimension(0:vector_size-1), device :: vector_output_gpu
real(kind=8) :: timer1,timer2,duration
integer :: streams(0:stream_count-1)
integer :: i,begin,end,size
logical :: success
integer :: error

! CPU code !

call random_seed()

allocate(vector_input_cpu(0:vector_size-1))
allocate(vector_output_cpu(0:vector_size-1))

! initialize cpu input vector to 0 !
do i=0,vector_size-1
  call random_number(vector_input_cpu(i))
end do
! initialize cpu output vector to 0 !
do i=0,vector_size-1
  vector_output_cpu(i)=0;
end do

! create streams !
do i=0,stream_count
  error = cudaStreamCreate(streams(i))
end do

call cpu_time(timer2)

do i=0,stream_count-1
  begin=vector_size*i/stream_count
  end=vector_size*(i+1)/stream_count
  size=end-begin
  ! copy input vector to gpu !
  error = cudaMemcpyAsync(vector_input_gpu(begin:end-1), &
                          vector_input_cpu(begin:end-1),size,streams(i))
end do

do i=0,stream_count-1
  begin=vector_size*i/stream_count
  end=vector_size*(i+1)/stream_count
  size=end-begin
  ! cuda kernel call !
  call compute<<<30,32,0,streams(i)>>>(vector_input_gpu(begin:end-1), &
                                       vector_output_gpu(begin:end-1),size)
end do

do i=0,stream_count-1
  begin=vector_size*i/stream_count
  end=vector_size*(i+1)/stream_count
  size=end-begin
  ! copy output vector from gpu to cpu !
  error = cudaMemcpyAsync(vector_output_cpu(begin:end-1), &
                          vector_output_gpu(begin:end-1), &
                          size,streams(i))
end do

error = cudaThreadSynchronize()
call cpu_time(timer2)
duration=timer2-timer1

! free streams !
do i=0,stream_count
  error = cudaStreamDestroy(streams(i))
end do

print *,'duration=',duration,' seconds'

! verify results !
success=.true.
do i=0,vector_size-1
  if (abs(iterate(vector_input_cpu(i))-vector_output_cpu(i))>1e-5) then
    print *,'computation result is wrong: index=',i, &
        ' expected result=',iterate(vector_input_cpu(i)), &      
        'cuda result=', vector_output_cpu(i)
    success=.false.
    exit
  end if
end do
if (success) then
  print *,'computation result is correct.'
end if

end program exercise11
