#include <stdio.h>
#include <math.h>
#include <omp.h>
#include "cuda_utils.h"

// CUDA Kernel //

const int blocks=90;

void __global__ ScalarProduct1(float* a,float* b,float* result,int size) {
  __shared__ float sums[16];
  float sum=0;
  for(int i=threadIdx.x+blockIdx.x*blockDim.x; i<size; 
      i+=blockDim.x*gridDim.x) {
    sum+=a[i]*b[i];
  }
  __syncthreads();
  if (threadIdx.x<16)
    sums[threadIdx.x]=sum;
  __syncthreads();
  if (threadIdx.x>=16)
    sums[threadIdx.x-16]+=sum;
  __syncthreads();
  if (threadIdx.x<8)
    sums[threadIdx.x]+=sums[threadIdx.x+8];
  __syncthreads();
  if (threadIdx.x<4)
    sums[threadIdx.x]+=sums[threadIdx.x+4];
  __syncthreads();
  if (threadIdx.x==0)
    result[blockIdx.x]=(sums[0]+sums[1])+(sums[2]+sums[3]);
}

void __global__ ScalarProduct2(float* result) {
  float sum=0;
  for(int i=0;i<blocks;i++) {
    sum+=result[i];
  }
  *result=sum;
}

// CPU code //

double computeScalarProduct(float* vector_a_cpu,float* vector_b_cpu,int size) {

  // allocate memory on gpu //
  float* vector_a_gpu;
  float* vector_b_gpu;
  float* result_gpu;
  cu_verify(cudaMalloc((void**)&vector_a_gpu,size*sizeof(float)));
  cu_verify(cudaMalloc((void**)&vector_b_gpu,size*sizeof(float)));
  cu_verify(cudaMalloc((void**)&result_gpu,blocks*sizeof(float)));

  // copy input vector from cpu to gpu //
  cu_verify(cudaMemcpy(vector_a_gpu,vector_a_cpu,size*sizeof(float),
                       cudaMemcpyHostToDevice));
  cu_verify(cudaMemcpy(vector_b_gpu,vector_b_cpu,size*sizeof(float),
                       cudaMemcpyHostToDevice));

  Timer timer;
  initTimer(&timer);

  // cuda kernel call //
  cu_verify_void((ScalarProduct1<<<blocks,32>>>(vector_a_gpu,vector_b_gpu,
                                                result_gpu,size)));
  cu_verify_void((ScalarProduct2<<<1,1>>>(result_gpu)));

  cu_verify(cudaThreadSynchronize());
  double duration = getTimer(&timer);

  int device;
  cu_verify(cudaGetDevice(&device));
  printf("computation on GPU %i: duration=%le seconds, "
         "performance = %le FLOPS\n",device,
         duration,2.*size/duration);

  float result_cpu = 0;

  // copy result from gpu to cpu //
  cu_verify(cudaMemcpy(&result_cpu,result_gpu,sizeof(float),
                       cudaMemcpyDeviceToHost));

  // free gpu memory //
  cu_verify(cudaFree(vector_a_gpu));
  cu_verify(cudaFree(vector_b_gpu));
  cu_verify(cudaFree(result_gpu));

  return result_cpu;
}

int main() {

  // define vector size //
  const int vector_size = 10000;

  // allocate memory on cpu //
  float* vector_a_cpu;
  float* vector_b_cpu;

  cu_verify(cudaSetDevice(0));
  cu_verify(cudaHostAlloc((void**)&vector_a_cpu,vector_size*sizeof(float),
			  cudaHostAllocPortable));
  cu_verify(cudaHostAlloc((void**)&vector_b_cpu,vector_size*sizeof(float),
			  cudaHostAllocPortable));

  float result;

  // initialize vectors //
  for(int i=0;i<vector_size;i++) {
    vector_a_cpu[i]=1.f/(1.f+i);
    vector_b_cpu[i]=1.f+i;
  }

  int first_size=vector_size/2;
  int second_size=vector_size-first_size;
  float sums[2];

  #pragma omp parallel
  {
    int thread=omp_get_thread_num();
    if (thread==0) {
      printf("starting computation on GPU 0\n");
      sums[0]=computeScalarProduct(vector_a_cpu,vector_b_cpu,first_size);
    }
    if (thread==1) {
      printf("starting computation on GPU 1\n");
      cu_verify(cudaSetDevice(1));
      sums[1]=computeScalarProduct(vector_a_cpu+first_size,
                                   vector_b_cpu+first_size,second_size);
    }
  }
  result=sums[0]+sums[1];

  // verify result //
  float sum=0;
  for(int i=0;i<vector_size;i++) {
    sum+=vector_a_cpu[i]*vector_b_cpu[i];
  }
  if (fabsf(sum-result)>1e-5) {
    printf("computation result is wrong: expected=%g cuda result=%g\n",
           sum,result);
  } else {
    printf("computation result is correct.\n");
  }

  // free cpu memory //
  cu_verify(cudaFreeHost(vector_a_cpu));
  cu_verify(cudaFreeHost(vector_b_cpu));

  // exit //
  return 0;
}
